import pytest
from pyspark.sql import SparkSession
import numpy as np
from pyspark.ml.linalg import Vectors
import torch.nn as nn
import torch
from sparktorch.inference import create_spark_torch_model
from sparktorch.util import serialize_torch_obj, serialize_torch_obj_lazy
from sparktorch.torch_distributed import SparkTorch
from sparktorch.tests.simple_net import Net, AutoEncoder, ClassificationNet, NetworkWithParameters


@pytest.fixture()
def spark():
    return (SparkSession.builder
            .master('local[2]')
            .appName('sparktorch')
            .getOrCreate())


@pytest.fixture()
def data(spark):
    dat = [(1.0, Vectors.dense(np.random.normal(0,1,10))) for _ in range(0, 200)]
    dat2 = [(0.0, Vectors.dense(np.random.normal(2,1,10))) for _ in range(0, 200)]
    dat.extend(dat2)
    return spark.createDataFrame(dat, ["label", "features"]).repartition(2)


@pytest.fixture()
def sequential_model():
    model = torch.nn.Sequential(
        nn.Linear(10, 20),
        nn.ReLU(),
        nn.Linear(20, 1)
    )
    return serialize_torch_obj(
        model, nn.MSELoss(), torch.optim.Adam, lr=0.001
    )


@pytest.fixture()
def lazy_model():
    model = serialize_torch_obj_lazy(
        Net, nn.MSELoss, torch.optim.Adam, optimizer_params={'lr': 0.001}
    )
    return model


@pytest.fixture()
def general_model():
    model = serialize_torch_obj(
        Net(), nn.MSELoss(), torch.optim.Adam, lr=0.001
    )
    return model


@pytest.fixture()
def network_with_params():
    return serialize_torch_obj_lazy(
        NetworkWithParameters,
        nn.MSELoss,
        torch.optim.Adam,
        optimizer_params={'lr': 0.001},
        model_parameters={'param': 40}
    )


def test_early_stopping_async(data, general_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=general_model,
        iters=25,
        verbose=1,
        earlyStopPatience=2
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]


def test_model_parameters(data, network_with_params):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=network_with_params,
        verbose=1,
        iters=5
    ).fit(data)

    py_model = stm.getPytorchModel()
    assert py_model.fc1 is not None
    assert py_model.fc2 is not None


def test_inference(lazy_model, data):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=lazy_model,
        verbose=1,
        iters=10
    ).fit(data)

    first_res = stm.transform(data).take(1)

    res = stm.getPytorchModel()
    spark_model = create_spark_torch_model(
        res,
        'features',
        'predictions'
    )

    res = spark_model.transform(data).take(1)
    assert first_res == res


def test_lazy(lazy_model, data):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=lazy_model,
        verbose=1,
        iters=5
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]
    assert type(res[0]['predictions']) is float


def test_simple_sequential(data, sequential_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=sequential_model,
        verbose=1,
        iters=5
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]
    assert type(res[0]['predictions']) is float


def test_simple_torch_module(data, general_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=general_model,
        iters=5,
        verbose=1
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]
    assert type(res[0]['predictions']) is float


def test_barrier(data, general_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=general_model,
        iters=5,
        verbose=1,
        partitions=2,
        useBarrier=True
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]


def test_autoencoder(data):
    model = serialize_torch_obj(
        AutoEncoder(), nn.MSELoss(), torch.optim.Adam, lr=0.001
    )

    stm = SparkTorch(
        inputCol='features',
        predictionCol='predictions',
        torchObj=model,
        iters=5,
        verbose=1,
        partitions=2,
        useVectorOut=True
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]
    assert len(res[0]['predictions']) == 10


def test_classification(data):
    model = serialize_torch_obj(
        ClassificationNet(), nn.CrossEntropyLoss(), torch.optim.Adam, lr=0.001
    )

    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=model,
        iters=5,
        verbose=1,
        partitions=2
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]


def test_mini_batch(data, general_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=general_model,
        iters=10,
        verbose=1,
        partitions=2,
        miniBatch=5,
        acquireLock=True
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]


def test_cpu_device(data, general_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=general_model,
        iters=10,
        verbose=1,
        partitions=2,
        miniBatch=5,
        acquireLock=True,
        device='cpu'
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]


def test_validation_pct(data, general_model):
    stm = SparkTorch(
        inputCol='features',
        labelCol='label',
        predictionCol='predictions',
        torchObj=general_model,
        iters=10,
        verbose=1,
        partitions=2,
        validationPct=0.25
    ).fit(data)

    res = stm.transform(data).take(1)
    assert 'predictions' in res[0]




"""
The MIT License

Copyright 2019 Derek Miller

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import codecs
import collections
import json
from typing import Any, Dict, List, Type, Tuple

import dill
import numpy as np
import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer

TorchObj = collections.namedtuple(
    'TorchObj', ['model', 'criterion', 'optimizer', 'optimizer_params', 'is_lazy', 'model_parameters']
)

DataObj = collections.namedtuple('DataObj', ['x_train', 'y_train', 'x_val', 'y_val'])


def torch_encoder(obj) -> str:
    """
    Encodes Torch Object or anything related
    :param obj: any object
    :return: decoded object
    """
    return codecs.encode(
        dill.dumps(obj), "base64"
    ).decode()


def torch_decoder(model_ser: str) -> Any:
    """
    Decodes the torch object or model
    :param model_ser: Serialized object
    :return: the object
    """
    return dill.loads(codecs.decode(model_ser.encode(), "base64"))


def handle_features(data: List[DataObj], validation_pct: float = 0.0) -> DataObj:
    """
    Processes features and converts them to torch tensors.

    :param data: The initial data is numpy vectors
    :param validation_pct: Percentage to use for a validation set
    :return: A DataObj with tensors

    """
    x_train = []
    y_train = []

    for feature in data:
        if feature.y_train is not None:
            if type(feature.y_train) is int or type(feature.y_train) is float:
                y_train.append([feature.y_train])
            else:
                y_train.append(feature.y_train)

        x_train.append(feature.x_train)

    if len(x_train) == 0:
        return DataObj(x_train=None, y_train=None, x_val=None, y_val=None)

    if validation_pct > 0:
        val_amt = int(len(x_train) * validation_pct)
        val_idxs = np.random.choice(len(x_train), val_amt, replace=False).tolist()
        train_idxs = list(set(val_idxs).symmetric_difference(set([i for i in range(len(x_train))])))

        full_x = torch.from_numpy(np.stack(x_train)).float()
        full_y = torch.from_numpy(np.asarray(y_train)).float() if y_train else None

        x_val = full_x[val_idxs]
        y_val = full_y[val_idxs] if full_y is not None else None

        x_train = full_x[train_idxs]
        y_train = full_y[train_idxs] if full_y is not None else None

        return DataObj(x_train=x_train, y_train=y_train, x_val=x_val, y_val=y_val)

    x_train = torch.from_numpy(np.stack(x_train)).float()
    y_train = torch.from_numpy(np.asarray(y_train)).float() if y_train else None

    return DataObj(x_train=x_train, y_train=y_train, x_val=None, y_val=None)


def load_base_torch(torch_obj: str) -> Tuple:
    """
    Loads the base torch object from json.
    :param torch_obj: The json torch object
    :return: a tuple of the torch object and the shapes of the parameters.
    """
    obj = json.loads(torch_obj)
    return obj['torch_obj'], obj['shapes']


def load_torch_model(torch_obj: str, from_json: bool = False) -> TorchObj:
    """
    Loads the torch object. If it is from json, it will load the json body before handling the request.

    :param torch_obj: The torch object as a serialized string
    :param from_json: Determines whether we should load from json
    :return: A loaded model torch object
    """
    if from_json:
        torch_obj, _ = load_base_torch(torch_obj)

    loaded = torch_decoder(torch_obj)
    if loaded.is_lazy:
        model = loaded.model(**loaded.model_parameters) if loaded.model_parameters else loaded.model()
        return TorchObj(
            model=model,
            criterion=loaded.criterion(),
            optimizer=load_optimizer(loaded.optimizer, loaded.optimizer_params, model),
            optimizer_params=loaded.optimizer_params,
            is_lazy=False,
            model_parameters=None
        )

    model = loaded.model
    optimizer = load_optimizer(loaded.optimizer, loaded.optimizer_params, model)
    return TorchObj(
        model=model,
        criterion=loaded.criterion,
        optimizer=optimizer,
        optimizer_params=loaded.optimizer_params,
        is_lazy=False,
        model_parameters=None
    )


def serialize_torch_obj_lazy(
    model: Type[nn.Module],
    criterion: Type[Any],
    optimizer: Type[Optimizer],
    optimizer_params: Dict = None,
    model_parameters: Dict = None
) -> str:
    """
    Lazily serializes a torch object.
    :param model: The network's class that you want to run.
    :param criterion: The class of the criterion.
    :param optimizer: The class of the optimizer.
    :param optimizer_params: The optimizer parameters as a dictionary.
    :param model_parameters: The model parameters as a dictionary.
    :return: a serialized json string of the torch model parameters.
    """
    tmp_mod = model(**model_parameters) if model_parameters else model()
    params = [list(ps.shape) for ps in tmp_mod.parameters()]
    model_encoded = torch_encoder(
        TorchObj(
            model=model,
            criterion=criterion,
            optimizer=optimizer,
            optimizer_params=optimizer_params,
            is_lazy=True,
            model_parameters=model_parameters
        )
    )
    return json.dumps({
        'torch_obj': model_encoded,
        'shapes': params
    })


def serialize_torch_obj(
        model: nn.Module,
        criterion: Any,
        optimizer: Type[Optimizer],
        **kwargs
) -> str:
    model_encoded = torch_encoder(
        TorchObj(
            model=model,
            criterion=criterion,
            optimizer=optimizer,
            optimizer_params=kwargs,
            is_lazy=False,
            model_parameters=None
        )
    )
    return json.dumps({
        'torch_obj': model_encoded,
        'shapes': [list(ps.shape) for ps in model.parameters()]
    })


def load_optimizer(optimizer: Type[Optimizer], params: Dict, model: nn.Module) -> Optimizer:
    if params:
        return optimizer(model.parameters(), **params)

    return optimizer(model.parameters())



from sparktorch.util import serialize_torch_obj, serialize_torch_obj_lazy
from sparktorch.torch_distributed import SparkTorch
from sparktorch.pipeline_util import PysparkPipelineWrapper
from sparktorch.inference import create_spark_torch_model



import torch.nn as nn
import torch.nn.functional as F


class Net(nn.Module):

    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(10, 20)
        self.fc2 = nn.Linear(20, 1)

    def forward(self, x):
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)
        return x


class AutoEncoder(nn.Module):

    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(10, 5)
        self.bottle = nn.Linear(5, 2)
        self.fc2 = nn.Linear(2, 5)
        self.out = nn.Linear(5, 10)

    def forward(self, x):
        x = self.fc1(x)
        x = F.relu(x)
        x = self.bottle(x)
        x = F.relu(x)
        x = self.fc2(x)
        x = F.relu(x)
        x = self.out(x)
        return x


class ClassificationNet(nn.Module):

    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(10, 20)
        self.fc2 = nn.Linear(20, 2)

    def forward(self, x):
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


class NetworkWithParameters(nn.Module):

    def __init__(self, param):
        super().__init__()
        self.fc1 = nn.Linear(10, param)
        self.fc2 = nn.Linear(param, 1)

    def forward(self, x):
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)
        return x



"""
The MIT License

Copyright 2019 Derek Miller

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import os
from typing import Dict, List, Union, Callable
from uuid import uuid4

import numpy as np
import torch
import torch.distributed as dist
from pyspark.rdd import PipelinedRDD
from pyspark.rdd import RDD
from pyspark.taskcontext import BarrierTaskContext

from sparktorch.early_stopper import EarlyStopping
from sparktorch.util import handle_features, load_torch_model, DataObj, load_base_torch


def get_available_port_port(address: str, context: BarrierTaskContext) -> int:
    port = ""
    if context.partitionId() == 0:
        try:
            import socket

            sock = socket.socket()
            sock.bind((address, 0))
            port = sock.getsockname()[1]
        except socket.error:
            pass
    available_port = context.allGather(str(port))[0]
    if not available_port:
        raise RuntimeError("Failed to find free port for distributed training.")

    return int(available_port)


def mapPartitionsWithIndex(rdd: RDD, f: Callable, preservesPartitioning: bool = False) -> PipelinedRDD:
    """
    Maps Partitions based off of the index.

    :param rdd: A valid rdd.
    :param f: Callable function with iterable.
    :param preservesPartitioning: Whether or not to preserve partitioning

    :return: A pipeline rdd.
    """
    return PipelinedRDD(rdd, f, preservesPartitioning, isFromBarrier=True)


def handle_model(
    index: int,
    data: List[DataObj],
    torch_obj: Union[str, List],
    iters: int = 1000,
    early_stop_patience: int = -1,
    verbose: int = 1,
    mini_batch: int = -1,
    validation_pct: float = 0,
    device: str = 'cpu',
    compile_mode: str = None,
) -> List[Dict]:
    """
    Runs the training of pytorch model, utilizing the distributed package.

    :param index: Partition index. Used for registering.
    :param data: The data from the partition
    :param torch_obj: The torch object string. Needs serialized
    :param iters: The iterations for training
    :param verbose: whether to log the loss or not.
    :param mini_batch: Mini batch for training
    :param validation_pct: Validation percentage.
    :param device: The pytorch device to use for training. cpu/cuda
    :param early_stop_patience: Amount of patient for early stopping. -1 means don't use early stopping.
    :param compile_mode: Torch compile option for 2.0 only.

    :return: A list of the model state dictionary.
    """
    # If a process has already been setup on the machine, kill it.
    if dist.is_initialized():
        dist.destroy_process_group()

    context = BarrierTaskContext.get()

    addrs = [e.address.split(":")[0] for e in context.getTaskInfos()]
    world_size = len(addrs)

    # Set up the distributed server.
    os.environ['MASTER_ADDR'] = str(addrs[0])
    os.environ['MASTER_PORT'] = str(get_available_port_port(addrs[0], context))
    os.environ['WORLD_SIZE'] = str(len(addrs))
    os.environ['NODE_RANK'] = str(context.partitionId())
    os.environ['RANK'] = str(context.partitionId())

    dist.init_process_group('gloo', rank=index, world_size=world_size)

    torch_obj = load_torch_model(torch_obj)

    # Loaded the model
    model = torch_obj.model.to(device)

    if compile is not None:
        torch.compile(compile_mode)

    model.train()
    criterion = torch_obj.criterion
    optimizer = torch_obj.optimizer

    # Set up early stopping
    es = EarlyStopping(patience=early_stop_patience)
    should_stop = torch.zeros(1)
    has_early_stop = early_stop_patience > 0

    partition_id = str(uuid4())

    # Process the data. Converts to x_train, y_train, x_val, y_val
    data_obj = handle_features(data, validation_pct)

    # Passes all of the data
    x_train = data_obj.x_train.to(device)
    y_train = data_obj.y_train.to(device) if data_obj.y_train is not None else x_train
    x_val = data_obj.x_val.to(device) if data_obj.x_val is not None else None
    y_val = data_obj.y_val.to(device) if data_obj.y_val is not None else x_val

    for i in range(iters):

        optimizer.zero_grad()

        # utilize minibatch
        if 0 < mini_batch < len(data_obj.x_train):
            idxs = np.random.choice(len(data_obj.x_train), mini_batch, replace=False).tolist()
            x_train = data_obj.x_train[idxs]
            y_train = data_obj.y_train[idxs]

        y_pred = model(x_train)

        try:
            loss = criterion(y_pred, y_train)
        except RuntimeError as e:
            # utilized when loss need a long label
            y_train = torch.flatten(y_train.long())
            loss = criterion(y_pred, y_train)

        loss_v = loss.item()

        # Process validation loss
        val_loss = None
        val_loss_v = None
        if x_val is not None:
            pred_val = model(x_val)

            try:
                val_loss = criterion(pred_val, y_val)
            except RuntimeError as e:
                y_val = torch.flatten(y_val.long())
                val_loss = criterion(pred_val, y_val)

            val_loss_v = val_loss.item()

        # Calculate gradients
        loss.backward()

        # Distributed part of training.
        for param in model.parameters():
            dist.all_reduce(param.grad.data, op=torch.distributed.ReduceOp.SUM)
            param.grad.data /= world_size

        # Processes the early stop work
        loss_distributed = None
        if has_early_stop:
            loss_to_use = val_loss if val_loss is not None else loss

            dist.all_reduce(loss_to_use, op=torch.distributed.ReduceOp.SUM)
            loss_distributed = loss_to_use.item() / world_size
            stop = es.step(loss_distributed)
            if stop:
                should_stop = should_stop + 1.0

            dist.all_reduce(should_stop, op=torch.distributed.ReduceOp.SUM)
            if should_stop.item() > 0:
                break

        optimizer.step()

        if verbose:
            print(f"Partition: {partition_id}. Iteration: {i}. Distributed Loss: {loss_distributed} "
                  f"Partition Training Loss: {loss_v}, "
                  f"Partition Validation Loss: {val_loss_v}")

    return [model.state_dict()]


def train_distributed(
    rdd: RDD,
    torch_obj: str,
    iters: int = 10,
    partition_shuffles: int = 1,
    verbose: int = 1,
    mini_batch: int = -1,
    validation_pct: float = 0.0,
    device: str = 'cpu',
    early_stop_patience: int = -1,
    compile_mode: str = None,
) -> Dict:
    """
    Entry point to train the model in distributed fashion.

    :param rdd: The rdd of data to run on the model.
    :param torch_obj: The torch object as a string that includes the model and param shapes.
    :param iters: Number of iterations for training.
    :param partition_shuffles: Number of partition shuffles.
    :param verbose: Verbosity of logs.
    :param mini_batch: Mini batch for each iteration of training.
    :param validation_pct: How many items to validate.
    :param device: pytorch device.
    :param early_stop_patience: amount of patience for early stopping.
    :param compile_mode: torch 2.0 compile mode.

    :return: The train dict.
    """
    torch_loaded, params = load_base_torch(torch_obj)

    state_dict = None
    for i in range(partition_shuffles):

        # Run model with barrier execution mode.
        state_dict = mapPartitionsWithIndex(
            rdd, lambda i, x: handle_model(
                i,
                x,
                torch_obj=torch_loaded,
                iters=iters,
                verbose=verbose,
                mini_batch=mini_batch,
                validation_pct=validation_pct,
                device=device,
                early_stop_patience=int(early_stop_patience + 0),
                compile_mode=compile_mode,
            )
        ).collect()

        if partition_shuffles - i > 1:
            num_partitions = rdd.getNumPartitions()
            rdd = rdd.repartition(num_partitions)

    return state_dict[0]



"""
Implementation is from here: https://gist.github.com/stefanonardo/693d96ceb2f531fa05db530f3e21517d by stefanonardo
"""

import numpy as np


class EarlyStopping(object):

    def __init__(self, mode='min', min_delta=0, patience=10, percentage=False):
        self.mode = mode
        self.min_delta = min_delta
        self.patience = patience
        self.best = None
        self.num_bad_epochs = 0
        self.is_better = None
        self._init_is_better(mode, min_delta, percentage)

        if patience == 0:
            self.is_better = lambda a, b: True
            self.step = lambda a: False

    def step(self, metrics):
        if self.best is None:
            self.best = metrics
            return False

        if np.isnan(metrics):
            return True

        if self.is_better(metrics, self.best):
            self.num_bad_epochs = 0
            self.best = metrics
        else:
            self.num_bad_epochs += 1

        if self.num_bad_epochs >= self.patience:
            return True

        return False

    def _init_is_better(self, mode, min_delta, percentage):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if not percentage:
            if mode == 'min':
                self.is_better = lambda a, best: a < best - min_delta
            if mode == 'max':
                self.is_better = lambda a, best: a > best + min_delta
        else:
            if mode == 'min':
                self.is_better = lambda a, best: a < best - (
                        best * min_delta / 100)
            if mode == 'max':
                self.is_better = lambda a, best: a > best + (
                        best * min_delta / 100)



import codecs

import dill
import torch.nn as nn
from pyspark.ml.pipeline import PipelineModel

from sparktorch.torch_distributed import SparkTorchModel


def convert_to_serialized_torch(network: nn.Module) -> str:
    """
    Converts an existing torch network to a serialized string.

    :param network: a nn.Module that you want to serialize
    :return: Returns the serialized torch model.
    """
    return codecs.encode(dill.dumps(network), "base64").decode()


def create_spark_torch_model(
    network: nn.Module,
    inputCol: str = 'features',
    predictionCol: str = 'predicted',
    useVectorOut: bool = False
) -> SparkTorchModel:
    """
    Creates a spark SparkTorchModel from an already trained network. Useful for running inference on large datasets.

    :param network: an already trained network
    :param inputCol: The spark dataframe input column
    :param predictionCol: The spark dataframe prediction columns
    :param useVectorOut: Determines whether the output should return a spark vector
    :return: Returns a SparkTorchModel
    """
    return SparkTorchModel(
        inputCol=inputCol,
        predictionCol=predictionCol,
        modStr=convert_to_serialized_torch(network),
        useVectorOut=useVectorOut
    )


def attach_pytorch_model_to_pipeline(
    network: nn.Module,
    pipeline_model: PipelineModel,
    inputCol: str = 'features',
    predictionCol: str = 'predicted',
    useVectorOut: bool = False
) -> PipelineModel:
    """
    Attaches a pytorch model to an existing pyspark pipeline.

    :param network: Pytorch Network
    :param pipeline_model: An existing spark pipeline model (This is a fitted pipeline)
    :param inputCol: The input column to the dataframe for the pytorch network
    :param predictionCol: The prediction column.
    :param useVectorOut: option to use a vector output.
    :return: a spark PipelineModel
    """
    spark_model = create_spark_torch_model(network, inputCol, predictionCol, useVectorOut)
    return PipelineModel(stages=[pipeline_model, spark_model])
